# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
import watchdog.events


import asyncio
import threading


class BatchingWatchDogHandler(watchdog.events.FileSystemEventHandler):
    def __init__(self, reload, extension=".toml"):
        self.lock = threading.Lock()
        self.timer = None
        self.reload = reload
        self.extension = extension

    def on_moved(self, event):
        return self.on_modified(event)

    def on_created(self, event):
        return self.on_modified(event)
    
    def on_deleted(self, event):
        return self.on_modified(event)

    def on_modified(self, event):
        if event.is_directory or not event.src_path.endswith(self.extension):
            return

        with self.lock:
            def reload():
                with self.lock:
                    if self.timer is not None:
                        self.timer.cancel()
                    self.timer = None
                    asyncio.run(self.reload())

            if self.timer is None:
                self.timer = threading.Timer(1.0, reload)
                self.timer.start()


# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.api.endpoints import run_code
from aiconsole.api.endpoints import projects
from fastapi import APIRouter

from aiconsole.api.endpoints import (
    chats,
    commands_history,
    profile,
    image,
    agents,
    ping,
    ws,
    materials,
    project_settings,
    check_key,
)


app_router = APIRouter()

app_router.include_router(ping.router)
app_router.include_router(image.router)
app_router.include_router(check_key.router)
app_router.include_router(run_code.router)
app_router.include_router(profile.router)
app_router.include_router(chats.router, prefix="/api/chats")
app_router.include_router(materials.router, prefix="/api/materials")
app_router.include_router(agents.router, prefix="/api/agents")
app_router.include_router(projects.router, prefix="/api/projects")
app_router.include_router(project_settings.router, prefix="/api/settings")
app_router.include_router(commands_history.router)
app_router.include_router(ws.router)



from aiconsole.core.assets.materials.material import MaterialContentType


def get_material_content_name(content_type: MaterialContentType):
    if content_type == "static_text":
        return "Note"
    elif content_type == "dynamic_text":
        return "Dynamic Note"
    elif content_type == "api":
        return "Python API"
    else:
        return "Material"



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

from aiconsole.core.analysis.agents_to_choose_from import agents_to_choose_from


def create_agents_str() -> str:
    """
    Randomization of agents is done because LLMs have a tendency to overfit to the first few examples.
    """

    # Forced agents if available or enabled agents otherwise
    possible_agent_choices = agents_to_choose_from()

    new_line = "\n"

    random_agents = new_line.join(
        [f"* {c.id} - {c.usage}" for c in random.sample(possible_agent_choices, len(possible_agent_choices))]
    )

    return random_agents



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.assets.asset import Asset
import tomlkit
from aiconsole.core.assets.agents.agent import Agent
from aiconsole.core.assets.fs.load_asset_from_fs import load_asset_from_fs
from aiconsole.core.assets.materials.material import Material, MaterialContentType
from aiconsole.core.project.paths import get_project_assets_directory


async def save_asset_to_fs(asset: Asset):
    path = get_project_assets_directory(asset.type)

    try:
        current_version = (await load_asset_from_fs(asset.type, asset.id)).version
    except KeyError:
        current_version = "0.0.1"

    # Parse version number
    current_version = current_version.split(".")

    # Increment version number
    current_version[-1] = str(int(current_version[-1]) + 1)

    # Join version number
    asset.version = ".".join(current_version)

    # Save to .toml file
    with (path / f"{asset.id}.toml").open("w") as file:
        # FIXME: preserve formatting and comments in the file using tomlkit

        # Ignore None values in model_dump
        model_dump = asset.model_dump()
        for key in list(model_dump.keys()):
            if model_dump[key] is None:
                del model_dump[key]

        def make_sure_starts_and_ends_with_newline(s: str):
            if not s.startswith("\n"):
                s = "\n" + s

            if not s.endswith("\n"):
                s = s + "\n"

            return s

        doc = tomlkit.document()
        doc.append("name", tomlkit.string(asset.name))
        doc.append("version", tomlkit.string(asset.version))
        doc.append("usage", tomlkit.string(asset.usage))
        doc.append("usage_examples", tomlkit.item(asset.usage_examples))
        doc.append("default_status", tomlkit.string(asset.default_status))

        if isinstance(asset, Material):
            doc.append("content_type", tomlkit.string(asset.content_type))

            {
                MaterialContentType.STATIC_TEXT: lambda: doc.append(
                    "content_static_text",
                    tomlkit.string(
                        make_sure_starts_and_ends_with_newline(asset.content_static_text),
                        multiline=True,
                    ),
                ),
                MaterialContentType.DYNAMIC_TEXT: lambda: doc.append(
                    "content_dynamic_text",
                    tomlkit.string(
                        make_sure_starts_and_ends_with_newline(asset.content_dynamic_text),
                        multiline=True,
                    ),
                ),
                MaterialContentType.API: lambda: doc.append(
                    "content_api",
                    tomlkit.string(
                        make_sure_starts_and_ends_with_newline(asset.content_api),
                        multiline=True,
                    ),
                ),
            }[asset.content_type]()

        if isinstance(asset, Agent):
            if asset.id == "user":
                raise Exception("Cannot save agent with id 'user'.")

            doc.append("system", tomlkit.string(asset.system))
            doc.append("gpt_mode", tomlkit.string(asset.gpt_mode))
            doc.append("execution_mode", tomlkit.string(asset.execution_mode))

        file.write(doc.as_string())

    return asset



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
from pathlib import Path
from typing import Any

import litellm
import tomlkit
import tomlkit.container
import tomlkit.exceptions
from aiconsole.api.websockets.outgoing_messages import SettingsWSMessage
from aiconsole.core.assets.asset import AssetStatus, AssetType
from aiconsole.core.project import project
from aiconsole.core.project.paths import get_project_directory
from aiconsole.core.project.project import is_project_initialized
from aiconsole.utils.BatchingWatchDogHandler import BatchingWatchDogHandler
from aiconsole.utils.recursive_merge import recursive_merge
from appdirs import user_config_dir
from pydantic import BaseModel
from tomlkit import TOMLDocument
from watchdog.observers import Observer

_log = logging.getLogger(__name__)


class PartialSettingsData(BaseModel):
    code_autorun: bool | None = None
    openai_api_key: str | None = None
    materials: dict[str, AssetStatus] = {}
    materials_to_reset: list[str] = []
    agents: dict[str, AssetStatus] = {}
    agents_to_reset: list[str] = []
    to_global: bool = False


class PartialSettingsAndToGlobal(PartialSettingsData):
    to_global: bool = False


class SettingsData(BaseModel):
    code_autorun: bool = False
    openai_api_key: str | None = None
    materials: dict[str, AssetStatus] = {}
    agents: dict[str, AssetStatus] = {}


def _load_from_path(file_path: Path) -> dict[str, Any]:
    with file_path.open() as file:
        document = tomlkit.loads(file.read())

    return dict(document)


def _set_openai_api_key_environment(settings: SettingsData) -> None:
    openai_api_key = settings.openai_api_key

    # This is so we make sure that it's overiden and does not env variables etc
    litellm.openai_key = openai_api_key or "invalid key"


class Settings:
    def __init__(self, project_path: Path | None = None):
        self._suppress_notification_until = None
        self._settings = SettingsData()

        self._global_settings_file_path = Path(user_config_dir("aiconsole")) / "settings.toml"

        if project_path:
            self._project_settings_file_path = project_path / "settings.toml"
        else:
            self._project_settings_file_path = None

        self._observer = Observer()

        self._global_settings_file_path.parent.mkdir(parents=True, exist_ok=True)
        self._observer.schedule(
            BatchingWatchDogHandler(self.reload, self._global_settings_file_path.name),
            str(self._global_settings_file_path.parent),
            recursive=True,
        )

        if self._project_settings_file_path:
            self._project_settings_file_path.parent.mkdir(parents=True, exist_ok=True)
            self._observer.schedule(
                BatchingWatchDogHandler(self.reload, self._project_settings_file_path.name),
                str(self._project_settings_file_path.parent),
                recursive=True,
            )

        self._observer.start()

    def model_dump(self) -> dict[str, Any]:
        return self._settings.model_dump()

    def stop(self):
        self._observer.stop()

    async def reload(self, initial: bool = False):
        self._settings = await self.__load()
        await SettingsWSMessage(
            initial=initial
            or not (
                not self._suppress_notification_until or self._suppress_notification_until < datetime.datetime.now()
            )
        ).send_to_all()
        self._suppress_notification_until = None

    def get_asset_status(self, asset_type: AssetType, id: str) -> AssetStatus:
        s = self._settings

        if asset_type == AssetType.MATERIAL:
            if id in s.materials:
                return s.materials[id]
            asset = project.get_project_materials().get_asset(id)
            default_status = asset.default_status if asset else AssetStatus.ENABLED
            return default_status
        elif asset_type == AssetType.AGENT:
            if id in s.agents:
                return s.agents[id]
            asset = project.get_project_agents().get_asset(id)
            default_status = asset.default_status if asset else AssetStatus.ENABLED
            return default_status

        else:
            raise ValueError(f"Unknown asset type {asset_type}")

    def rename_asset(self, asset_type: AssetType, old_id: str, new_id: str):
        if asset_type == AssetType.MATERIAL:
            partial_settings = PartialSettingsData(
                materials_to_reset=[old_id], materials={new_id: self.get_asset_status(asset_type, old_id)}
            )
        elif asset_type == AssetType.AGENT:
            partial_settings = PartialSettingsData(
                agents_to_reset=[old_id], agents={new_id: self.get_asset_status(asset_type, old_id)}
            )
        else:
            raise ValueError(f"Unknown asset type {asset_type}")

        self.save(partial_settings, to_global=True)

    def set_asset_status(self, asset_type: AssetType, id: str, status: AssetStatus, to_global: bool = False) -> None:
        if asset_type == AssetType.MATERIAL:
            self.save(PartialSettingsData(materials={id: status}), to_global=to_global)
        elif asset_type == AssetType.AGENT:
            self.save(PartialSettingsData(agents={id: status}), to_global=to_global)
        else:
            raise ValueError(f"Unknown asset type {asset_type}")

    def get_code_autorun(self) -> bool:
        return self._settings.code_autorun

    def set_code_autorun(self, code_autorun: bool, to_global: bool = False) -> None:
        self._settings.code_autorun = code_autorun
        self.save(PartialSettingsData(code_autorun=self._settings.code_autorun), to_global=to_global)

    async def __load(self) -> SettingsData:
        settings = {}
        paths = [self._global_settings_file_path, self._project_settings_file_path]

        for file_path in paths:
            if file_path and file_path.exists():
                settings = recursive_merge(settings, _load_from_path(file_path))

        forced_agents = [agent for agent, status in settings.get("agents", {}).items() if status == AssetStatus.FORCED]

        settings_materials = settings.get("materials", {})

        materials = {}
        for material, status in settings_materials.items():
            materials[material] = AssetStatus(status)

        agents = {}
        for agent, status in settings.get("agents", {}).items():
            agents[agent] = AssetStatus(status)

        settings_data = SettingsData(
            code_autorun=settings.get("settings", {}).get("code_autorun", False),
            openai_api_key=settings.get("settings", {}).get("openai_api_key", None),
            materials=materials,
            agents=agents,
        )

        # Enforce only one forced agent
        if len(forced_agents) > 1:
            _log.warning(f"More than one agent is forced: {forced_agents}")
            for agent in forced_agents[1:]:
                settings_data.agents[agent] = AssetStatus.ENABLED

        _set_openai_api_key_environment(settings_data)

        _log.info(f"Loaded settings")
        return settings_data

    @staticmethod
    def __get_tolmdocument_to_save(file_path: Path) -> TOMLDocument:
        if not file_path.exists():
            document = tomlkit.document()
            document["settings"] = tomlkit.table()
            document["materials"] = tomlkit.table()
            document["agents"] = tomlkit.table()
            return document

        with file_path.open() as file:
            document = tomlkit.loads(file.read())
            for section in ["settings", "materials", "agents"]:
                if section not in dict(document):
                    document[section] = tomlkit.table()

        return document

    def save(self, settings_data: PartialSettingsData, to_global: bool = False) -> None:
        if to_global:
            file_path = self._global_settings_file_path
        else:
            if not self._project_settings_file_path:
                raise ValueError("Cannnot save to project settings file, because project is not initialized")

            file_path = self._project_settings_file_path

        document = self.__get_tolmdocument_to_save(file_path)
        doc_settings: tomlkit.table.Table = document["settings"]
        doc_materials: tomlkit.table.Table = document["materials"]
        doc_agents: tomlkit.table.Table = document["agents"]

        if settings_data.code_autorun is not None:
            doc_settings["code_autorun"] = settings_data.code_autorun

        if settings_data.openai_api_key is not None:
            doc_settings["openai_api_key"] = settings_data.openai_api_key

        for material in settings_data.materials_to_reset:
            if material in doc_materials:
                del doc_materials[material]

        for agent in settings_data.agents_to_reset:
            if agent in doc_agents:
                del doc_agents[agent]

        for material in settings_data.materials:
            doc_materials[material] = settings_data.materials[material].value

        was_forced = False
        for agent in settings_data.agents:
            doc_agents[agent] = settings_data.agents[agent].value
            if settings_data.agents[agent] == AssetStatus.FORCED:
                was_forced = agent

        if was_forced:
            for agent in doc_agents:
                if agent != was_forced and doc_agents[agent] == AssetStatus.FORCED.value:
                    doc_agents[agent] = AssetStatus.ENABLED.value

        self._suppress_notification_until = datetime.datetime.now() + datetime.timedelta(seconds=30)

        file_path.parent.mkdir(parents=True, exist_ok=True)
        with file_path.open("w") as file:
            file.write(document.as_string())


async def init():
    global _settings
    _settings = Settings(get_project_directory() if is_project_initialized() else None)
    await _settings.reload()


def get_aiconsole_settings() -> Settings:
    return _settings


async def reload_settings(initial: bool = False):
    global _settings
    _settings.stop()
    _settings = Settings(get_project_directory() if is_project_initialized() else None)
    await _settings.reload(initial)



from aiconsole.core.assets.agents.agent import Agent
from aiconsole.core.assets.asset import AssetStatus
from aiconsole.core.project import project


from typing import cast


def agents_to_choose_from() -> list[Agent]:
    # Forced agents if available or enabled agents otherwise
    forced_agents = project.get_project_agents().assets_with_status(AssetStatus.FORCED)
    agents_to_choose_from = (
        forced_agents if forced_agents else project.get_project_agents().assets_with_status(AssetStatus.ENABLED)
    )
    agents_to_choose_from = cast(list[Agent], agents_to_choose_from)
    return agents_to_choose_from



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
from typing import Optional
from aiconsole.core.project.paths import get_project_directory
from aiconsole.core.project.project import change_project_directory, is_project_initialized

from fastapi import APIRouter
from pydantic import BaseModel

router = APIRouter()

_root = None


async def _ask_directory():
    from tkinter import Tk, filedialog

    global _root

    if not _root:
        _root = Tk()
    else:
        _root.deiconify()
    _root.withdraw()
    initial_dir = get_project_directory() if is_project_initialized() else os.getcwd()
    directory = filedialog.askdirectory(initialdir=initial_dir)

    # Check if the dialog was cancelled (directory is an empty string)
    if directory == "":
        return None
    else:
        return Path(directory)


class ChooseParams(BaseModel):
    directory: Optional[str] = None


@router.post("/is_project")
async def is_project(params: ChooseParams):
    directory = Path(params.directory) if params.directory else None

    if not directory:
        # Show a system select directory dialog
        directory = await _ask_directory()

    if not directory:
        return {"is_project": False, "path": None}

    is_project = directory.joinpath("materials").is_dir() or directory.joinpath("agents").is_dir()

    return {"is_project": is_project, "path": str(directory)}


@router.post("/choose")
async def choose_project(params: ChooseParams):
    directory = Path(params.directory) if params.directory else None

    if directory:
        await change_project_directory(directory)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.analysis.create_agents_str import create_agents_str
from aiconsole.core.analysis.create_materials_str import create_materials_str
from aiconsole.core.analysis.gpt_analysis_function_step import gpt_analysis_function_step
from aiconsole.core.chat.types import Chat
from aiconsole.core.gpt.consts import GPTMode


async def variant_quality_single_shot(chat: Chat, analysis_request_id: str):
    """
    FIRST SPEED, THEN QUALITY
    """

    await gpt_analysis_function_step(
        analysis_request_id=analysis_request_id,
        chat=chat,
        gpt_mode=GPTMode.QUALITY,
        initial_system_prompt=f"""
You are a director of a multiple AI Agents, doing everything to help the user.
You have multiple AI Agents at your disposal, each with their own unique capabilities.
Your job is to delegate tasks to the agents, and make sure that the user gets the best experience possible.
Never perform a task that an agent can do, and never ask the user to do something that an agent can do.
Do not answer other agents when they ask the user for something, allow the user to respond.
Be proactive, and try to figure out how to help without troubling the user.
If you spot an error in the work of an agent, suggest curreting it to the agent.
If an agent struggles with completing a task, experiment with giving him different set of materials.
If there is no meaningful next step, don't select an agent!
Your agents can only do things immediatelly, don't ask them to do something in the future.
Don't write or repeat any code, you don't know how to code.
Materials are special files that contain instructions for agents, you can choose which materials a given agent will have available, they can only use a limited number due to token limitations.

# Agents
You have the following agents available to handle the next step of this conversation, it can be one of the following ids (if next step is for user to respond, it should be 'user'):
{create_agents_str()}


# Materials
A list of ids of materials that are needed to execute the task, make sure that the agent has a prioritised list of those materials to look at, agents are not able to read all of them nor change your choice:
{create_materials_str()}

""".strip(),
        last_system_prompt=f"""
Your job is to prepare a plan.

1. Assess the current situation in the conversation, and who should now respond, the user or an agent?
2. Establish a full plan to bring value to the user
3. Briefly describe what the next, atomic, simple step of this conversation is, it can be both an action by a single agent or waiting for user response.
4. Establish who should handle the next step, it can be one of the following ids:
{create_agents_str()}

5. Figure out and provide a list of ids of materials that are needed to execute the task, choose among the following ids:
{create_materials_str()}

Questions to ask yourself:
- Is there a next step phrased as a next step, and a task for a given agent?
- Is there a better next step for this task?
- Are there any missing materials that could be useful for this task, that this solution does not have?
- Are there any materials that are not needed for this task, that this solution has?
- Are the materials sorted in an order of importance?
- Are you not repeating previous tasks and activity and delegating it to the same agent? Don't expect to get different results if you do the same thing again.
- Is there a better agent for this task?
- Is there a better way to describe the task?
- Is there anything that that might be a next task do that the user might find valuable? Are you trying to figure out how to help without troubling the user.
- Has an agent made an error in the last messages of the current conversation? If so maybe try to correct it with a different task, different agent or a different set of materials?
- If you are stuck you may always ask one agent to provide an expert critique of the current situation.
- Is the next step and agent correlated and choosen apropriatelly?
- If the next step is on the user, make sure that is_users_turn is True
- Does the solution contain a task for an agent? or is it an answer to the user? it should always be phrased as a task, and the answer should be given by the agent, not the director.
- Is the next step atomic?
- Is the next step the next logical step in this conversation?
- The next step should be either a single action for a single agent or a waiting for user response. If it's the latter, the agent selected should be the 'user'.

Now make the plan
""".strip(),
        force_call=True,
    )



#
# MIT License
#
# Copyright (c) 2023 Killian Lucas
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#     ____                      ____      __                            __
#    / __ \____  ___  ____     /  _/___  / /____  _________  ________  / /____  _____
#   / / / / __ \/ _ \/ __ \    / // __ \/ __/ _ \/ ___/ __ \/ ___/ _ \/ __/ _ \/ ___/
#  / /_/ / /_/ /  __/ / / /  _/ // / / / /_/  __/ /  / /_/ / /  /  __/ /_/  __/ /
#  \____/ .___/\___/_/ /_/  /___/_/ /_/\__/\___/_/  / .___/_/   \___/\__/\___/_/
#      /_/                                         /_/
#
# This file has been taken from the wonderful project "open-interpreter" by Killian Lucas
# https://github.com/KillianLucas/open-interpreter
#

import os
from typing import TYPE_CHECKING, List


if TYPE_CHECKING:
    from aiconsole.core.assets.assets import Assets

from ..subprocess_code_interpreter import SubprocessCodeInterpreter


class AppleScript(SubprocessCodeInterpreter):
    file_extension = "applescript"
    proper_name = "AppleScript"

    def __init__(self):
        super().__init__()
        self.start_cmd = os.environ.get("SHELL", "/bin/zsh")

    def preprocess_code(self, code, materials: List["Assets"]):
        """
        Inserts an end_of_execution marker and adds active line indicators.
        """

        # Escape double quotes
        code = code.replace('"', r"\"")

        # Wrap in double quotes
        code = '"' + code + '"'

        # Prepend start command for AppleScript
        code = "osascript -e " + code

        # Append end of execution indicator
        code += '; echo "## end_of_execution ##"'

        return code

    def detect_end_of_execution(self, line):
        """
        Detects end of execution marker in the output.
        """
        return "## end_of_execution ##" in line



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from aiconsole.core.chat.types import Chat
from aiconsole.core.project.paths import get_history_directory


def save_chat_history(chat: Chat):
    history_directory = get_history_directory()
    file_path = history_directory / f"{chat.id}.json"

    if len(chat.message_groups) == 0:
        # delete instead
        if os.path.exists(file_path):
            os.remove(file_path)
    else:
        os.makedirs(history_directory, exist_ok=True)
        with open(file_path, "w") as f:
            json.dump(chat.model_dump(exclude={"id", "last_modified"}),
                      f,
                      indent=4)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import traceback
from typing import cast

from aiconsole.api.websockets.outgoing_messages import ErrorWSMessage
from aiconsole.core.assets.materials.material import Material
from aiconsole.core.chat.chat_outgoing_messages import (
    SequenceStage,
    UpdateToolCallOutputWSMessage,
)
from aiconsole.core.code_running.run_code import get_code_interpreter
from aiconsole.core.code_running.types import CodeToRun
from aiconsole.core.project import project
from aiconsole.utils.cancel_on_disconnect import cancelable_endpoint
from fastapi import APIRouter, Request

router = APIRouter()

_log = logging.getLogger(__name__)


@router.post("/chats/{chat_id}/run_code")
@cancelable_endpoint
async def run_code(request: Request, code_to_run: CodeToRun, chat_id: str):
    try:
        _log.debug("Running code: %s", code_to_run.code)

        await UpdateToolCallOutputWSMessage(stage=SequenceStage.START, id=code_to_run.tool_call_id).send_to_chat(
            chat_id
        )

        mats = [project.get_project_materials().get_asset(mid) for mid in code_to_run.materials_ids]
        mats = [cast(Material, mat) for mat in mats if mat is not None]

        # If the code starts with a !, that means a shell command
        if code_to_run.language == "python" and code_to_run.code.startswith("!"):
            code_to_run.language = "shell"
            code_to_run.code = code_to_run.code[1:]

        try:
            async for token in get_code_interpreter(code_to_run.language).run(
                code_to_run.code, chat_id, code_to_run.tool_call_id, mats
            ):
                await UpdateToolCallOutputWSMessage(
                    stage=SequenceStage.MIDDLE,
                    id=code_to_run.tool_call_id,
                    output_delta=token,
                ).send_to_chat(chat_id)
        except asyncio.CancelledError:
            get_code_interpreter(code_to_run.language).terminate()
            _log.info("Run cancelled")
        except Exception:
            await ErrorWSMessage(error=traceback.format_exc().strip()).send_to_chat(chat_id)
            await UpdateToolCallOutputWSMessage(
                stage=SequenceStage.MIDDLE,
                id=code_to_run.tool_call_id,
                output_delta=traceback.format_exc().strip(),
            ).send_to_chat(chat_id)
    except Exception as e:
        await ErrorWSMessage(error=str(e)).send_to_chat(chat_id)
        raise e
    finally:
        await UpdateToolCallOutputWSMessage(stage=SequenceStage.END, id=code_to_run.tool_call_id).send_to_chat(chat_id)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import List
from pydantic import BaseModel


class EditableObject(BaseModel):
    id: str
    name: str


class AssetLocation(str, Enum):
    AICONSOLE_CORE = "aiconsole"
    PROJECT_DIR = "project"


class AssetType(str, Enum):
    AGENT = "agent"
    MATERIAL = "material"


class AssetStatus(str, Enum):
    DISABLED = "disabled"
    ENABLED = "enabled"
    FORCED = "forced"


class Asset(EditableObject):
    version: str = "0.0.1"
    usage: str
    usage_examples: List[str]
    defined_in: AssetLocation
    type: AssetType
    default_status: AssetStatus = AssetStatus.ENABLED
    status: AssetStatus = AssetStatus.ENABLED
    override: bool



from aiconsole.api.utils.status_change_post_body import StatusChangePostBody
from aiconsole.core.assets.asset import AssetType
from aiconsole.core.settings.project_settings import get_aiconsole_settings
from fastapi import HTTPException
from fastapi.responses import JSONResponse


async def asset_status_change(asset_type: AssetType, asset_id: str, body: StatusChangePostBody):
    """
    Change the status of a agent.

    Args:
        agent_id (str): The ID of the agent.
        body (StatusChangePostBody): POST body, only with "status"

    Returns:
        JSONResponse: JSON response indicating the result.
    """
    try:
        get_aiconsole_settings().set_asset_status(
            asset_type, id=asset_id, status=body.status, to_global=body.to_global
        )
        return JSONResponse({"status": "ok"})
    except KeyError:
        raise HTTPException(status_code=404, detail="Agent not found")
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from aiconsole.core.chat.types import AICMessage, AICMessageGroup, Chat
from aiconsole.consts import FUNCTION_CALL_OUTPUT_LIMIT
from aiconsole.core.gpt.types import (
    GPTFunctionCall,
    GPTRequestMessage,
    GPTRequestTextMessage,
    GPTRequestToolMessage,
    GPTToolCall,
)


from typing import List


last_system_message = None


def convert_message(group: AICMessageGroup, message: AICMessage, is_last: bool) -> List[GPTRequestMessage]:
    global last_system_message

    result = []

    #
    # Augment the messages with system messages with meta data about which agent is speaking and what materials were available
    #

    if group.task:
        system_message = f"""
As a director I have assigned you ({group.agent_id}) and given you access to the following materials text: {", ".join(group.materials_ids) if group.materials_ids else "None"}.
""".strip()

        # Only provide a task for last message
        if is_last:
            system_message += "\n\nYour job: " + group.task

        if last_system_message != system_message:
            result.append(
                GPTRequestTextMessage(
                    role="system",
                    name="director",
                    content=system_message,
                )
            )
            last_system_message = system_message

    tool_calls = [
        GPTToolCall(
            id=tool_call.id,
            function=GPTFunctionCall(
                name=tool_call.language,
                arguments=json.dumps(
                    {
                        "code": tool_call.code,
                    }
                ),
            ),
        )
        for tool_call in message.tool_calls
    ]

    result.append(
        GPTRequestTextMessage(
            role=group.role,
            content=message.content,
            name=group.agent_id if group.agent_id != "user" else None,
            tool_calls=tool_calls or None,
        ),
    )

    for tool_call in message.tool_calls:
        tool_call_id = tool_call.id

        content = tool_call.output

        if content is None:
            result.append(
                GPTRequestToolMessage(
                    tool_call_id=tool_call_id,
                    content="Running...",
                )
            )
        else:
            if content == "":
                content = "No output"

            # Enforce limit on output length, and put info that it was truncated only if limit was reached, truncate so the last part remains (not the first)
            if len(content) > FUNCTION_CALL_OUTPUT_LIMIT:
                content = f"""
Output truncated to last {FUNCTION_CALL_OUTPUT_LIMIT} characters:

...
{content[-FUNCTION_CALL_OUTPUT_LIMIT:]}
""".strip()

            result.append(GPTRequestToolMessage(tool_call_id=tool_call_id, content=content))

    return result


def convert_messages(chat: Chat) -> List[GPTRequestMessage]:
    global last_system_message
    last_system_message = None

    messages: List[GPTRequestMessage] = []

    for message_group in chat.message_groups:
        is_last_group = message_group == chat.message_groups[-1]
        for message in message_group.messages:
            is_last = is_last_group and message == message_group.messages[-1]
            messages.extend(convert_message(message_group, message, is_last=is_last))

    return messages



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from aiconsole.consts import COMMANDS_HISTORY_JSON, HISTORY_LIMIT
from aiconsole.core.chat.types import Command
from aiconsole.core.project import project
from aiconsole.core.project.paths import get_aic_directory
from fastapi import APIRouter

router = APIRouter()


@router.get("/commands/history")
def get_history():
    file_path = os.path.join(get_aic_directory(), COMMANDS_HISTORY_JSON)

    if os.path.exists(file_path):
        with open(file_path, "r") as f:
            commands = json.load(f)
    else:
        commands = []

    return commands


@router.post("/commands/history")
def save_history(command: Command):
    """
    Saves the history of sent commands to <commands_history_dir>/<commands_history_json>
    """
    file_path = os.path.join(get_aic_directory(), COMMANDS_HISTORY_JSON)

    if os.path.exists(file_path):
        with open(file_path, "r") as f:
            commands = json.load(f)
    else:
        commands = []

    commands.append(command.command)

    # remove non unique but keep the order, (but reverse it to remove the oldest first)
    commands.reverse()
    commands = list(dict.fromkeys(commands))
    commands.reverse()
    commands = commands[-HISTORY_LIMIT:]

    with open(file_path, "w") as f:
        json.dump(commands, f)

    return commands



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

Connection manager for websockets. Keeps track of all active connections

"""

import logging
from fastapi import WebSocket
from typing import List
from aiconsole.api.websockets.outgoing_messages import BaseWSMessage

_log = logging.getLogger(__name__)
_active_connections: List["AICConnection"] = []


class AICConnection:
    _websocket: WebSocket
    chat_id: str = ""

    def __init__(self, websocket: WebSocket):
        self._websocket = websocket

    async def send(self, msg: BaseWSMessage):
        await self._websocket.send_json({"type": msg.get_type(), **msg.model_dump()})


async def connect(websocket: WebSocket):
    await websocket.accept()
    connection = AICConnection(websocket)
    _active_connections.append(connection)
    _log.info("Connected")
    return connection


def disconnect(connection: AICConnection):
    _active_connections.remove(connection)
    _log.info("Disconnected")


async def send_message_to_chat(chat_id: str, msg: BaseWSMessage):
    # _log.debug(f"Sending message to {chat_id}: {msg}")
    for connection in _active_connections:
        if connection.chat_id == chat_id:
            await connection.send(msg)


async def send_message_to_all(msg: BaseWSMessage):
    # _log.debug(f"Sending message to all: {msg}")
    for connection in _active_connections:
        await connection.send(msg)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from typing import cast

from aiconsole.api.websockets.outgoing_messages import ErrorWSMessage
from aiconsole.core.assets.agents.agent import Agent, ExecutionModeContext
from aiconsole.core.assets.materials.material import Material
from aiconsole.core.chat.types import ChatWithAgentAndMaterials
from aiconsole.core.execution_modes.interpreter import execution_mode_interpreter
from aiconsole.core.execution_modes.normal import execution_mode_normal
from aiconsole.core.assets.materials.content_evaluation_context import ContentEvaluationContext
from aiconsole.core.project import project
from aiconsole.utils.cancel_on_disconnect import cancelable_endpoint
from fastapi import APIRouter, Request
from pydantic import ValidationError

router = APIRouter()
_log = logging.getLogger(__name__)


@router.post("/{chat_id}/execute")
@cancelable_endpoint
async def execute(request: Request, chat: ChatWithAgentAndMaterials, chat_id):
    if chat_id != chat.id:
        raise ValidationError("Chat ID does not match")

    agent = cast(Agent, project.get_project_agents().get_asset(chat.agent_id))

    content_context = ContentEvaluationContext(
        chat=chat,
        agent=agent,
        gpt_mode=agent.gpt_mode,
        relevant_materials=[
            cast(Material, project.get_project_materials().get_asset(id)) for id in chat.relevant_materials_ids
        ],
    )

    rendered_materials = [await material.render(content_context) for material in content_context.relevant_materials]

    context = ExecutionModeContext(
        chat=chat,
        agent=agent,
        relevant_materials=rendered_materials,
        gpt_mode=agent.gpt_mode,
    )

    try:
        execution_modes = {
            "interpreter": execution_mode_interpreter,
            "normal": execution_mode_normal,
        }
        execution_mode = execution_modes[agent.execution_mode]

        await execution_mode(context)
    except asyncio.CancelledError:
        _log.warning("Cancelled execution_mode_interpreter")
    except Exception as e:
        await ErrorWSMessage(error=str(e)).send_to_chat(chat.id)
        raise e



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.gpt.consts import MODEL_DATA
import openai

cached_good_keys = set()


# Verify the OpenAI key has access to the required models
async def check_key(key: str) -> bool:
    if key in cached_good_keys:
        return True

    try:
        # set key
        openai.api_key = key
        models = openai.Model.list(key=key)["data"]  # type: ignore
        available_models = [model["id"] for model in models]
        needed_models = MODEL_DATA.keys()

        good = set(needed_models).issubset(set(available_models))

        if good:
            cached_good_keys.add(key)

        return good
    except Exception as e:
        return False



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import List, Optional, Union

from aiconsole.core.gpt.types import GPTChoice, GPTFunctionCall, GPTResponse, GPTResponseMessage, GPTRole, GPTToolCall
from pydantic import BaseModel


def _parse_partial_json(s: str) -> Union[dict, str]:
    try:
        return json.loads(s)
    except json.JSONDecodeError:
        pass

    closing_chars_stack = []
    escaped = False
    completed_string = []

    for char in s:
        if closing_chars_stack and closing_chars_stack[-1] == '"':
            if char == '"' and not escaped:
                closing_chars_stack.pop()
            elif char == "\\":
                escaped = not escaped
            elif char == "\n" and not escaped:
                char = "\\n"
            else:
                escaped = False
        else:
            if char == '"':
                closing_chars_stack.append('"')
            elif char == "{":
                closing_chars_stack.append("}")
            elif char == "[":
                closing_chars_stack.append("]")
            elif char in ["}", "]"] and closing_chars_stack:
                closing_chars_stack.pop()

        completed_string.append(char)

    completed_string.extend(reversed(closing_chars_stack))

    try:
        return json.loads("".join(completed_string))
    except json.JSONDecodeError:
        return s


class GPTPartialFunctionCall(BaseModel):
    name: str = ""
    arguments_builder: List[str] = []

    @property
    def arguments(self) -> Union[dict, str]:
        self.arguments_builder = ["".join(self.arguments_builder)]
        return _parse_partial_json(self.arguments_builder[0])


class GPTPartialToolsCall(BaseModel):
    id: str = ""
    type: str = ""
    function: GPTPartialFunctionCall = GPTPartialFunctionCall()


class GPTPartialMessage(BaseModel):
    role: Optional[GPTRole] = None
    content_builder: Union[List[str], None] = None

    tool_calls: list[GPTPartialToolsCall] = []
    name: Optional[str] = None

    @property
    def content(self):
        if not self.content_builder:
            return None

        self.content_builder = ["".join(self.content_builder)]
        return self.content_builder[0]


class GPTPartialChoice(BaseModel):
    index: int = 0
    message: GPTPartialMessage = GPTPartialMessage()
    role: str = ""
    finnish_reason: str = ""


class GPTPartialResponse(BaseModel):
    id: str = ""
    object: str = ""
    created: int = 0
    model: str = ""
    choices: List[GPTPartialChoice] = []

    def to_final_response(self):
        return GPTResponse(
            id=self.id,
            object=self.object,
            created=self.created,
            model=self.model,
            choices=[
                GPTChoice(
                    index=choice.index,
                    message=GPTResponseMessage(
                        role=choice.message.role or "user",  # ???
                        content=choice.message.content,
                        tool_calls=[
                            GPTToolCall(
                                id=tool_call.id,
                                type=tool_call.type,
                                function=GPTFunctionCall(
                                    name=tool_call.function.name, arguments=tool_call.function.arguments
                                ),
                            )
                            for tool_call in choice.message.tool_calls
                        ],
                        name=choice.message.name,
                    ),
                    finnish_reason=choice.finnish_reason,
                )
                for choice in self.choices
            ],
        )

    def apply_chunk(self, chunk: dict):
        if "id" in chunk:
            self.id = chunk["id"]

        if "error" in chunk:
            raise ValueError(chunk["error"])

        if "object" in chunk:
            self.object = chunk["object"]

        if "created" in chunk:
            self.created = chunk["created"]

        if "model" in chunk:
            self.model = chunk["model"]

        if "choices" in chunk:
            chunk_choices = chunk["choices"]

            for chunk_choice in chunk_choices:
                index = None

                if "index" in chunk_choice:
                    index = int(chunk_choice["index"])

                if index is None:
                    raise ValueError("index not found")

                if index >= len(self.choices):
                    self.choices.append(GPTPartialChoice())

                choice = self.choices[index]
                choice.index = index

                if "finish_reason" in chunk_choice:
                    choice.finnish_reason = chunk_choice["finish_reason"]

                if "role" in chunk_choice:
                    choice.role = chunk_choice["role"]

                message = choice.message

                if "delta" in chunk_choice:
                    chunk_delta = chunk_choice["delta"]

                    if "name" in chunk_delta:
                        message.name = chunk_delta["name"]

                    if "role" in chunk_delta:
                        message.role = chunk_delta["role"]

                    if "content" in chunk_delta and chunk_delta["content"] != None:
                        if message.content_builder is None:
                            message.content_builder = []

                        message.content_builder.append(chunk_delta["content"])

                    if "tool_calls" in chunk_delta:
                        chunk_tool_calls = chunk_delta["tool_calls"]

                        for tool_call in chunk_tool_calls:
                            chunk_tool_index: int = tool_call["index"]
                            chunk_tool_function = tool_call["function"]

                            if chunk_tool_function:
                                while len(message.tool_calls) < chunk_tool_index + 1:
                                    message.tool_calls.append(
                                        GPTPartialToolsCall(
                                            id=tool_call["id"],
                                        )
                                    )

                                if tool_call.get("type"):
                                    message.tool_calls[chunk_tool_index].type = tool_call.get("type")

                                if chunk_tool_function:
                                    if "name" in chunk_tool_function:
                                        message.tool_calls[chunk_tool_index].function.name = chunk_tool_function.get(
                                            "name"
                                        )

                                    if "arguments" in chunk_tool_function:
                                        message.tool_calls[chunk_tool_index].function.arguments_builder.append(
                                            chunk_tool_function.get("arguments")
                                        )



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from aiconsole.core.code_running.code_interpreters.base_code_interpreter import BaseCodeInterpreter
from aiconsole.core.code_running.code_interpreters.language_map import language_map


code_interpreters = {}


def get_code_interpreter(language) -> BaseCodeInterpreter:
    if language not in code_interpreters:
        # Case in-sensitive
        language = language.lower()

        try:
            code_interpreters[language] = language_map[language]()
        except KeyError:
            raise ValueError(f"Unknown or unsupported language: {language}")

    return code_interpreters[language]


def reset_code_interpreters():
    global code_interpreters

    for code_interpreter in code_interpreters.values():
        code_interpreter.terminate()

    code_interpreters = {}



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from aiconsole.api.websockets.connection_manager import AICConnection
from aiconsole.api.websockets.outgoing_messages import (
    InitialProjectStatusWSMessage,
    ProjectClosedWSMessage,
    ProjectLoadingWSMessage,
    ProjectOpenedWSMessage,
)
from aiconsole.core.assets.asset import AssetType
from aiconsole.core.code_running.run_code import reset_code_interpreters

if TYPE_CHECKING:
    from aiconsole.core.assets import assets


_materials: Optional["assets.Assets"] = None
_agents: Optional["assets.Assets"] = None
_project_initialized = False


async def _clear_project():
    global _materials
    global _agents
    global _project_initialized

    if _materials:
        _materials.stop()

    if _agents:
        _agents.stop()

    reset_code_interpreters()

    _materials = None
    _agents = None
    _project_initialized = False


async def send_project_init(connection: AICConnection):
    from aiconsole.core.project.paths import get_project_directory, get_project_name

    await InitialProjectStatusWSMessage(
        project_name=get_project_name() if is_project_initialized() else None,
        project_path=str(get_project_directory()) if is_project_initialized() else None,
    ).send_to_chat(connection.chat_id)


def get_project_materials() -> "assets.Assets":
    if not _materials:
        raise ValueError("Project materials are not initialized")
    return _materials


def get_project_agents() -> "assets.Assets":
    if not _agents:
        raise ValueError("Project agents are not initialized")
    return _agents


def is_project_initialized() -> bool:
    return _project_initialized


async def close_project():
    from aiconsole.core.settings.project_settings import reload_settings

    await _clear_project()

    await ProjectClosedWSMessage().send_to_all()

    await reload_settings(initial=True)


async def reinitialize_project():
    from aiconsole.core.assets import assets
    from aiconsole.core.project.paths import get_project_directory, get_project_name
    from aiconsole.core.recent_projects.recent_projects import add_to_recent_projects
    from aiconsole.core.settings.project_settings import reload_settings

    await ProjectLoadingWSMessage().send_to_all()

    global _materials
    global _agents
    global _project_initialized

    await _clear_project()

    _project_initialized = True

    project_dir = get_project_directory()

    await add_to_recent_projects(project_dir)

    _agents = assets.Assets(asset_type=AssetType.AGENT)
    _materials = assets.Assets(asset_type=AssetType.MATERIAL)

    await ProjectOpenedWSMessage(path=str(get_project_directory()), name=get_project_name()).send_to_all()

    await _materials.reload(initial=True)
    await _agents.reload(initial=True)
    await reload_settings(initial=True)


async def change_project_directory(path: Path):
    if not path.exists():
        raise ValueError(f"Path {path} does not exist")

    # Change cwd and import path
    os.chdir(path)
    sys.path[0] = str(path)

    await reinitialize_project()



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
from multiprocessing import Process, Manager

manager = Manager()
shared_dict = manager.dict()

def display_shared_objects(dictionary):
    # Keep in mind that the dictionary is shared between processes, the subprocess does not have logging set up
    for material in dictionary:
        print('Material:', material)

    # Assuming `Materials` class has a method `to_dict` which returns a dictionary representation of all materials.
    # materials = Materials()
    # shared_dict.update(materials.to_dict())




from aiconsole.core.assets.asset import Asset, AssetType
from aiconsole.core.project import project
from fastapi import HTTPException
from fastapi.responses import JSONResponse


async def asset_post(asset_type: AssetType, asset: Asset, old_asset_id: str):
    try:
        if asset_type == AssetType.AGENT:
            assets = project.get_project_agents()
        elif asset_type == AssetType.MATERIAL:
            assets = project.get_project_materials()
        else:
            raise ValueError(f"Invalid asset type: {asset_type}")

        return JSONResponse({"renamed": await assets.save_asset(asset, old_asset_id=old_asset_id, create=True)})
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


async def asset_patch(asset_type: AssetType, asset: Asset, old_asset_id: str):
    if asset.id != old_asset_id:
        raise HTTPException(status_code=400, detail="Asset ID mismatch")

    try:
        if asset_type == AssetType.AGENT:
            assets = project.get_project_agents()
        elif asset_type == AssetType.MATERIAL:
            assets = project.get_project_materials()
        else:
            raise ValueError(f"Invalid asset type: {asset_type}")

        return await assets.save_asset(asset, old_asset_id=old_asset_id, create=False)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

from aiconsole.core.project.paths import get_history_directory


def list_possible_historic_chat_ids(project_path: Path | None = None):
    history_directory = get_history_directory(project_path)
    if history_directory.exists() and history_directory.is_dir():
        entries = os.scandir(history_directory)

        files = [entry for entry in entries if entry.is_file()
                 and entry.name.endswith(".json")]
        # Sort the files based on modification time (descending order)
        files = sorted(files, key=lambda entry: os.path.getmtime(
            entry.path), reverse=True)

        return [file.name.split(".")[0] for file in files]
    else:
        return []


#
# MIT License
#
# Copyright (c) 2023 Killian Lucas
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#     ____                      ____      __                            __
#    / __ \____  ___  ____     /  _/___  / /____  _________  ________  / /____  _____
#   / / / / __ \/ _ \/ __ \    / // __ \/ __/ _ \/ ___/ __ \/ ___/ _ \/ __/ _ \/ ___/
#  / /_/ / /_/ /  __/ / / /  _/ // / / / /_/  __/ /  / /_/ / /  /  __/ /_/  __/ /
#  \____/ .___/\___/_/ /_/  /___/_/ /_/\__/\___/_/  / .___/_/   \___/\__/\___/_/
#      /_/                                         /_/
#
# This file has been taken and slighly modified from the wonderful project
# "open-interpreter" by Killian Lucas https://github.com/KillianLucas/open-interpreter
#

import subprocess
import threading
import queue
import time
import traceback
from typing import AsyncGenerator, Generator, List

from aiconsole.core.assets.materials.material import Material
from .base_code_interpreter import BaseCodeInterpreter
import logging

_log = logging.getLogger(__name__)


class SubprocessCodeInterpreter(BaseCodeInterpreter):
    def __init__(self):
        self.start_cmd = ""
        self.process = None
        self.debug_mode = False
        self.output_queue: "queue.Queue[str]" = queue.Queue()
        self.done = threading.Event()

    def detect_end_of_execution(self, line):
        return None

    def line_postprocessor(self, line):
        return line

    def preprocess_code(self, code, materials: List[Material]):
        """
        This needs to insert an end_of_execution marker of some kind,
        which can be detected by detect_end_of_execution.

        Optionally, add active line markers for detect_active_line.
        """
        return code

    def terminate(self):
        if self.process:
            self.process.terminate()
        else:
            raise Exception("Process not started")

    def start_process(self):
        if self.process:
            self.terminate()

        self.process = subprocess.Popen(
            self.start_cmd.split(),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            bufsize=0,
            universal_newlines=True,
        )
        threading.Thread(target=self.handle_stream_output, args=(self.process.stdout, False), daemon=True).start()
        threading.Thread(target=self.handle_stream_output, args=(self.process.stderr, True), daemon=True).start()

    async def run(
        self, code: str, chat_id: str, tool_call_id: str, materials: List[Material]
    ) -> AsyncGenerator[str, None]:
        retry_count = 0
        max_retries = 3

        # Setup
        try:
            code = self.preprocess_code(code, materials)
            if not self.process:
                self.start_process()
        except:
            yield traceback.format_exc()
            return

        while retry_count <= max_retries:
            if self.debug_mode:
                print(f"Running code:\n{code}\n---")

            self.done.clear()

            try:
                if not self.process or not self.process.stdin:
                    raise Exception("Process not started")

                self.process.stdin.write(code + "\n")
                self.process.stdin.flush()
                break
            except:
                if retry_count != 0:
                    # For UX, I like to hide this if it happens once. Obviously feels better to not see errors
                    # Most of the time it doesn't matter, but we should figure out why it happens frequently with:
                    # applescript
                    yield traceback.format_exc()
                    yield f"Retrying... ({retry_count}/{max_retries})"
                    yield "Restarting process."

                self.start_process()

                retry_count += 1
                if retry_count > max_retries:
                    yield "Maximum retries reached. Could not execute code.."
                    return

        while True:
            if not self.output_queue.empty():
                yield self.output_queue.get()
            else:
                time.sleep(0.1)
            try:
                output = self.output_queue.get(timeout=0.3)  # Waits for 0.3 seconds
                yield output
            except queue.Empty:
                # AIConsole Fix: Added proces.pool check to fix hanging
                if self.done.is_set() or (self.process and self.process.poll() is not None):
                    # Try to yank 3 more times from it... maybe there's something in there...
                    # (I don't know if this actually helps. Maybe we just need to yank 1 more time)

                    for _ in range(3):
                        if not self.output_queue.empty():
                            yield self.output_queue.get()
                        time.sleep(0.2)
                    break

    def handle_stream_output(self, stream, is_error_stream):
        for line in iter(stream.readline, ""):
            _log.debug(f"Received output line:\n{line}\n---")

            line = self.line_postprocessor(line)

            if line is None:
                continue  # `line = None` is the postprocessor's signal to discard completely

            if self.detect_end_of_execution(line):
                self.done.set()
            elif is_error_stream and "KeyboardInterrupt" in line:
                self.output_queue.put("KeyboardInterrupt")
                time.sleep(0.1)
                self.done.set()
            else:
                self.output_queue.put(line)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
from aiconsole.core.assets.asset import AssetType
from aiconsole.core.project.project import is_project_initialized
from aiconsole.utils.resource_to_path import resource_to_path

def get_project_assets_directory(asset_type: AssetType, project_path: Path | None = None):
    if not is_project_initialized() and not project_path:
        raise ValueError("Project settings are not initialized")

    return get_project_directory(project_path) / f"{asset_type.value}s"

def get_core_assets_directory(asset_type: AssetType):
    return resource_to_path(f"aiconsole.preinstalled.{asset_type.value}s")


def get_history_directory(project_path: Path | None = None):
    if not is_project_initialized() and not project_path:
        raise ValueError("Project settings are not initialized")
    return get_aic_directory(project_path) / "history"


def get_aic_directory(project_path: Path | None = None):
    if not is_project_initialized() and not project_path:
        raise ValueError("Project settings are not initialized")
    return get_project_directory(project_path) / ".aic"


def get_project_directory(project_path: Path | None = None):
    if not is_project_initialized() and not project_path:
        raise ValueError("Project settings are not initialized")

    if project_path:
        return project_path

    return Path(os.getcwd())


def get_credentials_directory(project_path: Path | None = None):
    if not is_project_initialized() and not project_path:
        raise ValueError("Project settings are not initialized")
    return get_aic_directory(project_path) / "credentials"


def get_project_name(project_path: Path | None = None):
    if not is_project_initialized() and not project_path:
        raise ValueError("Project settings are not initialized")
    return get_project_directory(project_path).name
